# THIS FILE IS GENERATED FROM SETUP_EXT
version = '0.8.2'
full_version = '0.8.2'
git_revision = '4ce881644c872e629358e31f37bc8d099ce42591'
is_released = True

msi_version = '0.8.2.843'

version_info = (0, 8, 2, 'final', 0)
